/*
 * top_21_23_xc.h
 *
 *  Created on: 12. mars 2021
 *      Author: teig
 */

#ifndef TOP_21_23_XC_H_
#define TOP_21_23_XC_H_

    #if ((TOPOLOGY == TOP_21_HAS_1_TILE_004_T0_4N_RUNS_FAIR) or \
         (TOPOLOGY == TOP_23_HAS_2_TILE_004_T0_8N_T1_8N_CHANENDS_OVERRUN))
    #else
        #error Not TOP_21 or TOP_23
    #endif

    /*
    _______ ____  _____    ___  __                   _   _______ ____  _____    ___  ____
    |__   __/ __ \|  __ \  |__ \/_ |                 | | |__   __/ __ \|  __ \  |__ \|___ \
       | | | |  | | |__) |    ) || |   __ _ _ __   __| |    | | | |  | | |__) |    ) | __) |
       | | | |  | |  ___/    / / | |  / _` | '_ \ / _` |    | | | |  | |  ___/    / / |__ <
       | | | |__| | |       / /_ | | | (_| | | | | (_| |    | | | |__| | |       / /_ ___) |
       |_|  \____/|_|      |____||_|  \__,_|_| |_|\__,_|    |_|  \____/|_|      |____|____/
    */

    typedef interface conn_to_server_if_t {
        void to_server (const temp_degC_t);
    } conn_to_server_if_t;

    typedef interface conn_to_client_if_t {
        void to_client (const temp_degC_t);
    } conn_to_client_if_t;


    [[combinable]]
    void Client_node_task_root (
            const  unsigned            iof_row,
            const  unsigned            iof_col,
            client conn_to_server_if_t i_conn_out_0,
            client conn_to_server_if_t i_conn_out_1,
            client conn_to_server_if_t i_conn_out_2,
            server conn_to_client_if_t i_conn_in_0,
            server conn_to_client_if_t i_conn_in_1,
            server conn_to_client_if_t i_conn_in_2,
            out buffered port:4        outP4_leds)
        {

        node_context_t context;
        const params_t params    = init_params ();
        temp_degC_t    temp_degC = init_temp (iof_row,iof_col, params);

        timer    tmr;
        time32_t time_ticks;
        bool     guard_do_send = true;
        unsigned from_server_cnt = NUM_CONNS_PER_NODE;

        const time32_t time_delay_ticks = ((iof_row==0) and (iof_col==0)) ? ROOT_WAIT_FOR_NEXT_ROUND_US : 0;

        unsigned leds = BOARD_LEDS_INIT;

        outP4_leds <: leds;

        init_wt      (iof_row, iof_col, context.wt, params);
        init_context (iof_row, iof_col, context);

        PRINT_BANNER_ROOT (context, CLIENT_ROOT_STR, "Client_node_task_root", temp_degC, TOP_STR);
        context.cycle_cnt++; // To get data set to become like TOP_20's

        tmr :> time_ticks;

        while (1) { // Client_node_task_root
            select {
                case guard_do_send => tmr when timerafter (time_ticks) :> time_ticks : {
                    // Swap LED
                    leds xor_eq BOARD_LED_MASK_GREEN_ONLY; // J1.7 CH1 XCORE-200-EXPLORER
                    outP4_leds <: leds;

                    PRINT_NL_PROGESS (context,CLIENT_STR,0,0);
                    i_conn_out_0.to_server(temp_degC);
                    PRINT_PROGESS (context,CLIENT_STR,0,1);
                    i_conn_out_1.to_server(temp_degC);
                    PRINT_PROGESS (context,CLIENT_STR,0,2);
                    i_conn_out_2.to_server(temp_degC);
                    PRINT_PROGESS (context,CLIENT_STR,0,3);
                    guard_do_send = false;
                    from_server_cnt = 0;
                } break;

                case i_conn_in_0.to_client (const temp_degC_t temp_degC_in) : {
                    context.temps_degC[0] = temp_degC_in;
                    PRINT_PROGESS (context,CLIENT_STR,1,0);

                    from_server_cnt++;
                    if (from_server_cnt == NUM_CONNS_PER_NODE) {
                        PRINT_PROGESS (context,CLIENT_STR,1,100);
                        temp_degC = calculate_new_temp_from_flow (context, temp_degC);
                        PRINT_VALUES (context, CLIENT_STR, temp_degC);
                        context.cycle_cnt++;

                        guard_do_send = true;
                        time_ticks = time_ticks + time_delay_ticks;
                    }
                } break;
                case i_conn_in_1.to_client (const temp_degC_t temp_degC_in) : {
                    context.temps_degC[1] = temp_degC_in;
                    PRINT_PROGESS (context,CLIENT_STR,1,1);

                    from_server_cnt++;
                    if (from_server_cnt == NUM_CONNS_PER_NODE) {
                        PRINT_PROGESS (context,CLIENT_STR,1,101);
                        temp_degC = calculate_new_temp_from_flow (context, temp_degC);
                        PRINT_VALUES (context, CLIENT_STR, temp_degC);
                        context.cycle_cnt++;

                        guard_do_send = true;
                        time_ticks = time_ticks + time_delay_ticks;
                    }
                } break;
                case i_conn_in_2.to_client (const temp_degC_t temp_degC_in) : {
                    context.temps_degC[2] = temp_degC_in;
                    PRINT_PROGESS (context,CLIENT_STR,1,2);

                    from_server_cnt++;
                    if (from_server_cnt == NUM_CONNS_PER_NODE) {
                        PRINT_PROGESS (context,CLIENT_STR,1,102);
                        temp_degC = calculate_new_temp_from_flow (context, temp_degC);
                        PRINT_VALUES (context, CLIENT_STR, temp_degC);
                        context.cycle_cnt++;

                        guard_do_send = true;
                        time_ticks = time_ticks + time_delay_ticks;
                    }
                } break;
            }
        }
    } // Client_node_task_root


    [[combinable]]
    void Client_node_task (
            const  unsigned            iof_row,
            const  unsigned            iof_col,
            client conn_to_server_if_t i_conn_out_0,
            client conn_to_server_if_t i_conn_out_1,
            client conn_to_server_if_t i_conn_out_2,
            server conn_to_client_if_t i_conn_in_0,
            server conn_to_client_if_t i_conn_in_1,
            server conn_to_client_if_t i_conn_in_2)
        {

        node_context_t context;
        const params_t params    = init_params ();
        temp_degC_t    temp_degC = init_temp (iof_row,iof_col, params);

        timer    tmr;
        time32_t time_ticks;
        bool     guard_do_send = true;
        unsigned from_server_cnt = NUM_CONNS_PER_NODE;

        const time32_t time_delay_ticks = ((iof_row==0) and (iof_col==0)) ? ROOT_WAIT_FOR_NEXT_ROUND_US : 0;
        //const time32_t time_delay_ticks = ROOT_WAIT_FOR_NEXT_ROUND_US;

        init_wt      (iof_row, iof_col, context.wt, params);
        init_context (iof_row, iof_col, context);

        PRINT_BANNER (context, CLIENT_STR, "Client_node_task", temp_degC);
        context.cycle_cnt++; // To get data set to become like TOP_20's

        tmr :> time_ticks;

        while (1) { // Client_node_task
            select {
                case guard_do_send => tmr when timerafter (time_ticks) :> time_ticks : {
                    PRINT_PROGESS (context,CLIENT_STR,0,0);
                    i_conn_out_0.to_server(temp_degC);
                    PRINT_PROGESS (context,CLIENT_STR,0,1);
                    i_conn_out_1.to_server(temp_degC);
                    PRINT_PROGESS (context,CLIENT_STR,0,2);
                    i_conn_out_2.to_server(temp_degC);
                    PRINT_PROGESS (context,CLIENT_STR,0,3);
                    guard_do_send = false;
                    from_server_cnt = 0;
                } break;

                case i_conn_in_0.to_client (const temp_degC_t temp_degC_in) : {
                    context.temps_degC[0] = temp_degC_in;
                    PRINT_PROGESS (context,CLIENT_STR,1,0);

                    from_server_cnt++;
                    if (from_server_cnt == NUM_CONNS_PER_NODE) {
                        PRINT_PROGESS (context,CLIENT_STR,1,100);
                        temp_degC = calculate_new_temp_from_flow (context, temp_degC);
                        PRINT_VALUES (context, CLIENT_STR, temp_degC);
                        context.cycle_cnt++;

                        guard_do_send = true;
                        time_ticks = time_ticks + time_delay_ticks;
                    }
                } break;
                case i_conn_in_1.to_client (const temp_degC_t temp_degC_in) : {
                    context.temps_degC[1] = temp_degC_in;
                    PRINT_PROGESS (context,CLIENT_STR,1,1);

                    from_server_cnt++;
                    if (from_server_cnt == NUM_CONNS_PER_NODE) {
                        PRINT_PROGESS (context,CLIENT_STR,1,101);
                        temp_degC = calculate_new_temp_from_flow (context, temp_degC);
                        PRINT_VALUES (context, CLIENT_STR, temp_degC);
                        context.cycle_cnt++;

                        guard_do_send = true;
                        time_ticks = time_ticks + time_delay_ticks;
                    }
                } break;
                case i_conn_in_2.to_client (const temp_degC_t temp_degC_in) : {
                    context.temps_degC[2] = temp_degC_in;
                    PRINT_PROGESS (context,CLIENT_STR,1,2);

                    from_server_cnt++;
                    if (from_server_cnt == NUM_CONNS_PER_NODE) {
                        PRINT_PROGESS (context,CLIENT_STR,1,102);
                        temp_degC = calculate_new_temp_from_flow (context, temp_degC);
                        PRINT_VALUES (context, CLIENT_STR, temp_degC);
                        context.cycle_cnt++;

                        guard_do_send = true;
                        time_ticks = time_ticks + time_delay_ticks;
                    }
                } break;
            }
        }
    } // Client_node_task


    [[combinable]]
    void Server_node_task (
            const  unsigned            iof_row,
            const  unsigned            iof_col,
            server conn_to_server_if_t i_conn_in_0,
            server conn_to_server_if_t i_conn_in_1,
            server conn_to_server_if_t i_conn_in_2,
            client conn_to_client_if_t i_conn_out_0,
            client conn_to_client_if_t i_conn_out_1,
            client conn_to_client_if_t i_conn_out_2)
        {

        node_context_t context;
        const params_t params        = init_params ();
        temp_degC_t    temp_degC     = init_temp (iof_row,iof_col, params);
        temp_degC_t    temp_degC_old = temp_degC;

        timer    tmr;
        time32_t time_ticks;
        bool     guard_do_send   = false;
        unsigned to_client_cnt   = 0;
        unsigned from_client_cnt = 0;

        init_wt      (iof_row, iof_col, context.wt, params);
        init_context (iof_row, iof_col, context);

        PRINT_BANNER (context, CLIENT_STR, "Server_node_task", temp_degC);
        context.cycle_cnt++;

        tmr :> time_ticks; // To get data set to become like TOP_20's

        while (1) { // Server_node_task
            select {
                case i_conn_in_0.to_server (const temp_degC_t temp_degC_in) : {
                    temp_degC_old = temp_degC;
                    context.temps_degC[0] = temp_degC_in;
                    PRINT_PROGESS (context,SERVER_STR,0,0);

                    from_client_cnt++;
                    if (from_client_cnt == NUM_CONNS_PER_NODE) {
                        PRINT_PROGESS (context,SERVER_STR,0,100);
                        temp_degC = calculate_new_temp_from_flow (context, temp_degC);
                        PRINT_VALUES (context, SERVER_STR, temp_degC);
                        to_client_cnt = 0;
                        from_client_cnt = 0;
                        guard_do_send = true;
                        tmr :> time_ticks; // immediately
                    }
                } break;
                case i_conn_in_1.to_server (const temp_degC_t temp_degC_in) : {
                    temp_degC_old = temp_degC;
                    context.temps_degC[1] = temp_degC_in;
                    PRINT_PROGESS (context,SERVER_STR,0,1);

                    from_client_cnt++;
                    if (from_client_cnt == NUM_CONNS_PER_NODE) {
                        PRINT_PROGESS (context,SERVER_STR,0,101);
                        temp_degC = calculate_new_temp_from_flow (context, temp_degC);
                        PRINT_VALUES (context, SERVER_STR, temp_degC);
                        to_client_cnt = 0;
                        from_client_cnt = 0;
                        guard_do_send = true;
                        tmr :> time_ticks; // immediately
                    }
                } break;
                case i_conn_in_2.to_server (const temp_degC_t temp_degC_in) : {
                    temp_degC_old = temp_degC;
                    context.temps_degC[2] = temp_degC_in;
                    PRINT_PROGESS (context,SERVER_STR,0,2);

                    from_client_cnt++;
                    if (from_client_cnt == NUM_CONNS_PER_NODE) {
                        PRINT_PROGESS (context,SERVER_STR,0,102);
                        temp_degC = calculate_new_temp_from_flow (context, temp_degC);
                        PRINT_VALUES (context, SERVER_STR, temp_degC);
                        to_client_cnt = 0;
                        from_client_cnt = 0;
                        guard_do_send = true;
                        tmr :> time_ticks; // immediately
                    }
                } break;
                case guard_do_send => tmr when timerafter (time_ticks) :> void : {
                    PRINT_PROGESS (context,SERVER_STR,1,to_client_cnt);
                    switch (to_client_cnt) {
                        case 2: {i_conn_out_2.to_client(temp_degC_old);} break;
                        case 1: {i_conn_out_1.to_client(temp_degC_old);} break;
                        case 0: {i_conn_out_0.to_client(temp_degC_old);} break;
                    }

                    to_client_cnt++;
                    if (to_client_cnt == NUM_CONNS_PER_NODE) {
                        guard_do_send = false;
                        context.cycle_cnt++;
                    } else {}

                    tmr :> time_ticks; // immediately
                } break;
            }
        }
    } // Server_node_task

    int main (void) {
        #if ((NUM_ROWS==2) and (NUM_COLS==2))
            // TOP_21_HAS_1_TILE_004_T0_4N_RUNS_FAIR
            /*
            Constraint check for tile[0]:
              Cores available:            8,   used:          4 .  OKAY
              Timers available:          10,   used:          4 .  OKAY
              Chanends available:        32,   used:         17 .  OKAY
              Memory available:       262144,   used:      37920 .  OKAY
                (Stack: 9532, Code: 25942, Data: 2446)
            Constraints checks PASSED.
            */
            #if (TOP_WARNINGS==1) // From "makefile"
                #warning TOP_STR "Top_21"
            #endif

            conn_to_server_if_t conn_tx [NUM_ROWS][NUM_PAIRS_PER_ROW][NUM_CONNS_PER_NODE]; // [ROW][PAIR][CONN] = [R][P][C]
            conn_to_client_if_t conn_rx [NUM_ROWS][NUM_PAIRS_PER_ROW][NUM_CONNS_PER_NODE]; // [ROW][PAIR][CONN] = [R][P][C]
            //                          [    2   ][         1       ][        3         ]
            par {
                #define VER_NS_X [1][0][1] // NS      North-South top-bottom
                #define HOR_1_EW [1][0][2] // EW      EastWest belt side-to-side
                #define HOR_1_A  [1][0][0] // [A,B,C] Between horisontal columns
                #define VER_01_X [0][0][1] // [X,Y,Z] Between vertical rows
                #define HOR_0_EW [0][0][2] // EW      EastWest belt side-to-side
                #define HOR_0_A  [0][0][0] // [A,B,C] Between horisontal columns

                par {
                    Client_node_task_root (0,0, conn_tx HOR_0_EW, conn_tx VER_NS_X, conn_tx HOR_0_A,
                                                conn_rx HOR_0_EW, conn_rx VER_NS_X, conn_rx HOR_0_A, outP4_leds);
                    Server_node_task      (0,1, conn_tx HOR_0_A,  conn_tx VER_01_X, conn_tx HOR_0_EW,
                                                conn_rx HOR_0_A,  conn_rx VER_01_X, conn_rx HOR_0_EW);
                }

                // [[combine]] // XMOS: uncomment and it comoiles, with this "xcc1: internal compiler error in ..lower_combined_pars.cpp, line 183"
                par {
                    Client_node_task      (1,0, conn_tx HOR_1_EW, conn_tx VER_01_X, conn_tx HOR_1_A,
                                                conn_rx HOR_1_EW, conn_rx VER_01_X, conn_rx HOR_1_A);
                    Server_node_task      (1,1, conn_tx HOR_1_A,  conn_tx VER_NS_X, conn_tx HOR_1_EW,
                                                conn_rx HOR_1_A,  conn_rx VER_NS_X, conn_rx HOR_1_EW);
                }
            }
        #elif ((NUM_ROWS==4) and (NUM_COLS==4))
            // TOP_23_HAS_2_TILE_004_T0_8N_T1_8N_CHANENDS_OVERRUN
            /*
            Constraint check for tile[0]:
              Cores available:            8,   used:          8 .  OKAY
              Timers available:          10,   used:          8 .  OKAY
              Chanends available:        32,   used:         37 .  FAILED
              Memory available:       262144,   used:      53464 .  OKAY
                (Stack: 22836, Code: 27724, Data: 2904)
            Error: Constraints check FAILED for tile[0].
            Constraint check for tile[1]:
              Cores available:            8,   used:          8 .  OKAY
              Timers available:          10,   used:          8 .  OKAY
              Chanends available:        32,   used:         37 .  FAILED
              Memory available:       262144,   used:      52024 .  OKAY
                (Stack: 23316, Code: 26180, Data: 2528)
            Error: Constraints check FAILED for tile[1].
            */
            #if (TOP_WARNINGS==1) // From "makefile"
                #warning TOP_STR "Top_23"
            #endif

            conn_to_server_if_t conn_tx [NUM_ROWS][NUM_PAIRS_PER_ROW][NUM_CONNS_PER_NODE]; // [ROW][PAIR][CONN] = [R][P][C]
            conn_to_client_if_t conn_rx [NUM_ROWS][NUM_PAIRS_PER_ROW][NUM_CONNS_PER_NODE]; // [ROW][PAIR][CONN] = [R][P][C]
            //                          [    4   ][         2       ][        3         ]
            par {
                #define VER_NS_X [3][0][1] // NS      North-South top-bottom
                #define VER_NS_Y [3][1][1] // NS      North-South top-bottom

                #define HOR_3_EW [3][1][2] // EW      EastWest belt side-to-side
                #define HOR_3_A  [3][0][0] // [A,B,C] Between horisontal columns
                #define HOR_3_B  [3][0][2] // [A,B,C] Between horisontal columns
                #define HOR_3_C  [3][1][0] // [A,B,C] Between horisontal columns

                #define VER_23_X [2][0][1] // [X,Y,Z] Between vertical rows
                #define VER_23_Y [2][1][1] // [X,Y,Z] Between vertical rows

                #define HOR_2_EW [2][1][2] // EW      EastWest belt side-to-side
                #define HOR_2_A  [2][0][0] // [A,B,C] Between horisontal columns
                #define HOR_2_B  [2][0][2] // [A,B,C] Between horisontal columns
                #define HOR_2_C  [2][1][0] // [A,B,C] Between horisontal columns

                #define VER_12_X [1][0][1] // [X,Y,Z] Between vertical rows
                #define VER_12_Y [1][1][1] // [X,Y,Z] Between vertical rows

                #define HOR_1_EW [1][1][2] // EW      EastWest belt side-to-side
                #define HOR_1_A  [1][0][0] // [A,B,C] Between horisontal columns
                #define HOR_1_B  [1][0][2] // [A,B,C] Between horisontal columns
                #define HOR_1_C  [1][1][0] // [A,B,C] Between horisontal columns

                #define VER_01_X [0][0][1] // [X,Y,Z] Between vertical rows
                #define VER_01_Y [0][1][1] // [X,Y,Z] Between vertical rows

                #define HOR_0_EW [0][1][2] // EW      EastWest belt side-to-side
                #define HOR_0_A  [0][0][0] // [A,B,C] Between horisontal columns
                #define HOR_0_B  [0][0][2] // [A,B,C] Between horisontal columns
                #define HOR_0_C  [0][1][0] // [A,B,C] Between horisontal columns

                on tile[0]:
                // [[combine]] "xcc1: internal compiler error in ../lower_combined_pars.cpp, line 183"
                // But works in https://www.teigfam.net/oyvind/home/technology/202-kode24/
                par {
                    Client_node_task_root (0,0, conn_tx HOR_0_EW, conn_tx VER_NS_X, conn_tx HOR_0_A,
                                                conn_rx HOR_0_EW, conn_rx VER_NS_X, conn_rx HOR_0_A, outP4_leds);
                    Server_node_task      (0,1, conn_tx HOR_0_A,  conn_tx VER_01_X, conn_tx HOR_0_B,
                                                conn_rx HOR_0_A,  conn_rx VER_01_X, conn_rx HOR_0_B);
                    Client_node_task      (0,2, conn_tx HOR_0_B,  conn_tx VER_NS_Y, conn_tx HOR_0_C,
                                                conn_rx HOR_0_B,  conn_rx VER_NS_Y, conn_rx HOR_0_C);
                    Server_node_task      (0,3, conn_tx HOR_0_C,  conn_tx VER_01_Y, conn_tx HOR_0_EW,
                                                conn_rx HOR_0_C,  conn_rx VER_01_Y, conn_rx HOR_0_EW);

                    Client_node_task      (1,0, conn_tx HOR_1_EW, conn_tx VER_01_X, conn_tx HOR_1_A,
                                                conn_rx HOR_1_EW, conn_rx VER_01_X, conn_rx HOR_1_A);
                    Server_node_task      (1,1, conn_tx HOR_1_A,  conn_tx VER_12_X, conn_tx HOR_1_B,
                                                conn_rx HOR_1_A,  conn_rx VER_12_X, conn_rx HOR_1_B);
                    Client_node_task      (1,2, conn_tx HOR_1_B,  conn_tx VER_01_Y, conn_tx HOR_1_C,
                                                conn_rx HOR_1_B,  conn_rx VER_01_Y, conn_rx HOR_1_C);
                    Server_node_task      (1,3, conn_tx HOR_1_C,  conn_tx VER_12_Y, conn_tx HOR_1_EW,
                                                conn_rx HOR_1_C,  conn_rx VER_12_Y, conn_rx HOR_1_EW);
                }

                on tile[1]:
                // [[combine]] "xcc1: internal compiler error in ../lower_combined_pars.cpp, line 183"
                // But works in https://www.teigfam.net/oyvind/home/technology/202-kode24/
                par {
                    Client_node_task      (2,0, conn_tx HOR_2_EW, conn_tx VER_12_X, conn_tx HOR_2_A,
                                                conn_rx HOR_2_EW, conn_rx VER_12_X, conn_rx HOR_2_A);
                    Server_node_task      (2,1, conn_tx HOR_2_A,  conn_tx VER_23_X, conn_tx HOR_2_B,
                                                conn_rx HOR_2_A,  conn_rx VER_23_X, conn_rx HOR_2_B);
                    Client_node_task      (2,2, conn_tx HOR_2_B,  conn_tx VER_12_Y, conn_tx HOR_2_C,
                                                conn_rx HOR_2_B,  conn_rx VER_12_Y, conn_rx HOR_2_C);
                    Server_node_task      (2,3, conn_tx HOR_2_C,  conn_tx VER_23_Y, conn_tx HOR_2_EW,
                                                conn_rx HOR_2_C,  conn_rx VER_23_Y, conn_rx HOR_2_EW);

                    Client_node_task      (3,0, conn_tx HOR_3_EW, conn_tx VER_23_X, conn_tx HOR_3_A,
                                                conn_rx HOR_3_EW, conn_rx VER_23_X, conn_rx HOR_3_A);
                    Server_node_task      (3,1, conn_tx HOR_3_A,  conn_tx VER_NS_X, conn_tx HOR_3_B,
                                                conn_rx HOR_3_A,  conn_rx VER_NS_X, conn_rx HOR_3_B);
                    Client_node_task      (3,2, conn_tx HOR_3_B,  conn_tx VER_23_Y, conn_tx HOR_3_C,
                                                conn_rx HOR_3_B,  conn_rx VER_23_Y, conn_rx HOR_3_C);
                    Server_node_task      (3,3, conn_tx HOR_3_C,  conn_tx VER_NS_Y, conn_tx HOR_3_EW,
                                                conn_rx HOR_3_C,  conn_rx VER_NS_Y, conn_rx HOR_3_EW);
                }
            }
        #endif

        return 0;

    } // main

#endif /* TOP_21_23_XC_H_ */
