/*
 * top_24_25_xc.h
 *
 *  Created on: 12. mars 2021
 *      Author: teig
 */


#ifndef TOP_24_25_XC_H_
#define TOP_24_25_XC_H_
    /*
      _______ ____  _____    ___  _  _     ___  _____
     |__   __/ __ \|  __ \  |__ \| || |   |__ \| ____|
        | | | |  | | |__) |    ) | || |_     ) | |__
        | | | |  | |  ___/    / /|__   _|   / /|___ \
        | | | |__| | |       / /_   | |    / /_ ___) |
        |_|  \____/|_|      |____|  |_|   |____|____/
    */

    #if (TOPOLOGY == TOP_24_HAS_1_TILE_004_T0_2N_2CN_ETC_MAYBE_CONVERGING)
        // ok
    #elif (TOPOLOGY == TOP_25_HAS_1_TILE_016_T0_2N_14CN_NOT_CONVERGING)
        // ok
    #else
        #error Not Top_24 or 25
    #endif
    /*
     _                    _
    | |                  | |
    | |     _____   _____| |_   _
    | |    / _ \ \ / / _ \ | | | |
    | |___| (_) \ V /  __/ | |_| |
    |______\___/ \_/ \___|_|\__, |
                             __/ |
                            |___/
    */
    //-------------------------------------------------------------------------------------------------
    // TOP_20_HAS_2_TILE_004_T0_2N_BY3SUB_T1_2N_BY3SUB_RUNS_FAIR
    // FAIR and SYNCHRONIZE, but will not scale to more than 2*2
    // Not fair, does not synchronise, also see blog note
    // https://www.teigfam.net/oyvind/home/technology/217-my-xc-combined-combinable-notes/#all_combined

        typedef int chan_of_t; // ... chans don't have types. It's up to the user to ensure that sent and received are equal

    //-------------------------------------------------------------------------------------------------
    // TOP_21_HAS_1_TILE_004_T0_4N_RUNS_FAIR
    // but does not scale: TOP_23_HAS_2_TILE_004_T0_8N_T1_8N_CHANENDS_OVERRUN

        typedef interface conn_to_server_if_t {
            void to_server (const temp_degC_t);
        } conn_to_server_if_t;

        typedef interface conn_to_client_if_t {
            void to_client (const temp_degC_t);
        } conn_to_client_if_t;
    /*
     ____
    / __ \
    | |  | |_ __   ___ _ __     ___ __ _ ___  ___  ___
    | |  | | '_ \ / _ \ '_ \   / __/ _` / __|/ _ \/ __|
    | |__| | |_) |  __/ | | | | (_| (_| \__ \  __/\__ \
     \____/| .__/ \___|_| |_|  \___\__,_|___/\___||___/
         | |
         |_|
    */
    //-------------------------------------------------------------------------------------------------
    // TOP_01 .. TOP_15
    // Not fair, does not synchronise, also see blog note
    // But very [[combinable]] [[combine]] friendly - if they don't stop
    // https://www.teigfam.net/oyvind/home/technology/217-my-xc-combined-combinable-notes/#all_combined

    typedef interface conn_if_t_ {
       {unsigned, temp_degC_t} do_io_server (const temp_degC_t);
    } conn_if_t_; //

    //-------------------------------------------------------------------------------------------------
    // TOP_03A_HAS_2_TILE_016_NODE_T0_2N_AS_CC_6CN_T1_8CN_STOPS
    // Theoretically very nice, allows for asynchronousity and being in synch and fair, but probably
    // is not made for multi-clients. See https://www.xcore.com/viewtopic.php?f=26&t=8090

    typedef interface a_conn_if_t {
        [[guarded]]             void       do_io_server     (const unsigned value_to);
        [[notification]] slave  void       all_clients_seen (void);
        [[clears_notification]] {unsigned} get_result       (void);
    } a_conn_if_t;

    //-------------------------------------------------------------------------------------------------
    // TOP_03G_HAS_2_TILE_016_NODE_T0_2N_AS_CC_6CN_T1_8CN_NO_BUILD
    // This [[guarded]] protocol woule have solved the whole thing, but thec ompiler will not build the config

    typedef interface g_conn_if_t {
        [[guarded]] {unsigned} do_io_server (
                const unsigned value_to);
    } g_conn_if_t;

    /*
      _______ ____  _____    ___  _  _     ___  _____
     |__   __/ __ \|  __ \  |__ \| || |   |__ \| ____|
        | | | |  | | |__) |    ) | || |_     ) | |__
        | | | |  | |  ___/    / /|__   _|   / /|___ \
        | | | |__| | |       / /_   | |    / /_ ___) |
        |_|  \____/|_|      |____|  |_|   |____|____/
    */

    //-------------------------------------------------------------------------------------------------
    // TOP_24_HAS_1_TILE_004_T0_2N_2CN_ETC_MAYBE_CONVERGING
    // This is a CLIENT DRIVEN interface only, it therefore introduces pollling to poll_all_synched.
    // Client-driven with no [[notification]] etc. seems to be the only way to scale,
    // since it allows [[combinable]] [[combine]]

    typedef interface conn_if_t {
        temp_degC_t set_get (const temp_degC_t temp_degC);
        //
        // FAIR and SYNCHRONISED and COMBINABLE! Will some times imply more than one polling per round trip:
        all_clients_synched_t poll_all_synched (void);
    } conn_if_t;


    typedef enum client_state_e {
        SEND_DO_IO_SERVER,
        SYNCHRONIZE
    } client_state_e;

    //-------------------------------------------------------------------------------------------------
    // Client_node_task
    // This at first sends data to all three servers, then polls all three until all three servers
    // have been visited by all three clients. This makes this architecture "fair" and "synchronized".
    //

    [[combinable]]
    void Client_node_task (
            const  unsigned     iof_row,
            const  unsigned     iof_col,
            client conn_if_t    i_conn_0,
            client conn_if_t    i_conn_1,
            client conn_if_t    i_conn_2,
            out buffered port:4 ?outP4_leds_)
        {

        node_context_t context;
        const params_t params     = init_params ();
        temp_degC_t    temp_degC  = init_temp (iof_row,iof_col, params);
        unsigned       iof_client = 0;

        all_clients_synched_t all_in_synch[NUM_CONNS_PER_NODE] = {false, false, false};

        timer          tmr;
        time32_t       time_ticks;
        const bool     root         = is_root (iof_row, iof_col);
        client_state_e client_state = SEND_DO_IO_SERVER;

        unsigned leds = BOARD_LEDS_INIT;

        if (not isnull (outP4_leds_)) {
            outP4_leds_ <: leds;
        }

        init_wt      (iof_row, iof_col, context.wt, params);
        init_context (iof_row, iof_col, context);

        if (root) {
            PRINT_BANNER_ROOT (context, CLIENT_STR, "Client_node_task root", temp_degC, TOP_STR);
        } else {
            PRINT_BANNER (context, CLIENT_STR, "Client_node_task", temp_degC);
        }

        context.cycle_cnt++; // To get data set to become like TOP_20's

        tmr :> time_ticks;

        while (1) { // Client_node_task
            select {
                case tmr when timerafter (time_ticks) :> void : {
                    if (client_state == SEND_DO_IO_SERVER) {
                        switch (iof_client) {
                            case 0: {
                                PRINT_PROGESS (context,CLIENT_STR,0,0);
                                context.temps_degC[0] = i_conn_0.set_get (temp_degC);
                            } break;
                            case 1: {
                                 PRINT_PROGESS (context,CLIENT_STR,0,1);
                                 context.temps_degC[1] = i_conn_1.set_get (temp_degC);
                            } break;
                            case 2: {
                                 PRINT_PROGESS (context,CLIENT_STR,0,2);
                                 context.temps_degC[2] = i_conn_2.set_get (temp_degC);
                            } break;
                            // other: let the system crash
                        }

                        iof_client++;

                        if (iof_client == NUM_CONNS_PER_NODE) {
                            iof_client = 0;
                            temp_degC = calculate_new_temp_from_flow (context, temp_degC);

                            if (root) {
                                PRINT_VALUES_WITH_TIME_MOD (context, CLIENT_STR, tmr, temp_degC, DO_PRINT_VALUES_EVERY_CYCLE_CNT);
                            } else {
                                PRINT_VALUES_MOD (context, CLIENT_STR, temp_degC, DO_PRINT_VALUES_EVERY_CYCLE_CNT);
                            }

                            context.cycle_cnt++;

                            if (not isnull (outP4_leds_)) {
                                // Swap LED
                                leds xor_eq BOARD_LED_MASK_GREEN_ONLY; // J1.7 CH1 XCORE-200-EXPLORER
                                outP4_leds_ <: leds;
                            }

                            client_state = SYNCHRONIZE;
                            tmr :> time_ticks;
                            time_ticks += CLIENT_DELAY_UNTIL_NEXT_SYNCH_POLL_US * XS1_TIMER_MHZ;
                        } else {
                            tmr :> time_ticks;
                            time_ticks += CLIENT_DELAY_UNTIL_NEXT_THIRD_US * XS1_TIMER_MHZ;
                            // Allowing skew here, opposite of TOP_20_HAS_2_TILE_004_T0_2N_BY3SUB_T1_2N_BY3SUB_RUNS_FAIR,
                            // since the below tmr_synch would add unknown synch time anyhow
                            // TODO: unless we had that synch be encompassed by the outer tmr
                        }
                    } else if (client_state == SYNCHRONIZE) {
                        // With the 2*2 nodes case, THIS client may have two connections to the same server.
                        // However, we handle the general case

                        PRINT_NL_PROGESS (context,CLIENT_STR,1,0);
                        if (not all_in_synch[0]) {
                            all_in_synch[0] = i_conn_0.poll_all_synched ();
                        } else {}

                        PRINT_PROGESS (context,CLIENT_STR,1,1);
                        if (not all_in_synch[1]) {
                            all_in_synch[1] = i_conn_1.poll_all_synched ();
                        } else {}

                        PRINT_PROGESS (context,CLIENT_STR,1,2);
                        if (not all_in_synch[2]) {
                             all_in_synch[2] = i_conn_2.poll_all_synched ();
                        } else {}

                        PRINT_PROGESS (context,CLIENT_STR,1,200);


                        if (all_in_synch[0] and all_in_synch[1] and all_in_synch[2]) {
                            PRINT_PROGESS (context,CLIENT_STR,1,99);
                            for (unsigned ix=0; ix<NUM_CONNS_PER_NODE; ix++) {
                                all_in_synch[ix] = false;
                            }
                            iof_client = 0;
                            client_state = SEND_DO_IO_SERVER;
                            tmr :> time_ticks;
                            time_ticks += (ROOT_WAIT_FOR_NEXT_ROUND_US * (unsigned) root); // Delay for root, not for the other clients
                        } else {
                            tmr :> time_ticks;
                            time_ticks += (CLIENT_DELAY_UNTIL_NEXT_SYNCH_POLL_US * XS1_TIMER_MHZ);
                        }
                    }
                } break;
            }
        }
    } // Client_node_task


    void Handle_server_set_get (
            node_context_t     &context, // io
            unsigned           &from_client_set_get_cnt, // io
            all_clients_synched_t &all_synched, // io
            temp_degC_t        &temp_degC) // io
    {

        PRINT_PROGESS (context,SERVER_STR,0,0);
        from_client_set_get_cnt++;
        if (from_client_set_get_cnt == NUM_CONNS_PER_NODE) {
            from_client_set_get_cnt = 0;
            all_synched = true;
            //
            PRINT_PROGESS (context,SERVER_STR,0,100);
            temp_degC = calculate_new_temp_from_flow (context, temp_degC);
            PRINT_VALUES_MOD (context, SERVER_STR, temp_degC, DO_PRINT_VALUES_EVERY_CYCLE_CNT);
            context.cycle_cnt++;
        } else {}
    }

    void Handle_server_poll_all_clients_seen (
            all_clients_synched_t &all_synched, // io
            unsigned           &from_client_poll_all_synched_cnt, // io
            all_clients_synched_t &all_in_synch) // io
    {

        if (all_synched) {
            all_in_synch = true;
            from_client_poll_all_synched_cnt++;
            if (from_client_poll_all_synched_cnt == NUM_CONNS_PER_NODE) {
                from_client_poll_all_synched_cnt = 0;
                all_synched = false;
            }
        } else {
            all_in_synch = false;
        }
        PRINT_PROGESS (context,SERVER_STR,1,200+all_in_synch_return);
    }

    [[combinable]]
    void Server_node_task (
            const  unsigned  iof_row,
            const  unsigned  iof_col,
            server conn_if_t i_conn_0,
            server conn_if_t i_conn_1,
            server conn_if_t i_conn_2)
        {

        node_context_t context;
        const params_t     params                           = init_params ();
        temp_degC_t        temp_degC                        = init_temp (iof_row,iof_col, params);
        unsigned           from_client_set_get_cnt          = 0;
        unsigned           from_client_poll_all_synched_cnt = 0;
        all_clients_synched_t all_synched                      = false;

        init_wt      (iof_row, iof_col, context.wt, params);
        init_context (iof_row, iof_col, context);

        PRINT_BANNER (context, SERVER_STR, "Server_node_task", temp_degC);
        context.cycle_cnt++;

        while (1) { // Server_node_task
            // [[ordered]] not allowed in combinable task
            select {
                // -----------------------------------------------------------------------------
                // PHASE I: setting and delivering temperature, calculate when all clients seen
                // -----------------------------------------------------------------------------
                case i_conn_0.set_get (const temp_degC_t temp_degC_in) -> temp_degC_t temp_degC_return : {
                    temp_degC_return = temp_degC; // Always present value
                    context.temps_degC[0] = temp_degC_in;
                    Handle_server_set_get (context, from_client_set_get_cnt, all_synched, temp_degC);
                } break;
                case i_conn_1.set_get (const temp_degC_t temp_degC_in) -> temp_degC_t temp_degC_return : {
                    temp_degC_return = temp_degC; // Always present value
                    context.temps_degC[1] = temp_degC_in;
                    Handle_server_set_get (context, from_client_set_get_cnt, all_synched, temp_degC);
                } break;
                case i_conn_2.set_get (const temp_degC_t temp_degC_in) -> temp_degC_t temp_degC_return : {
                    temp_degC_return = temp_degC; // Always present value
                    context.temps_degC[2] = temp_degC_in;
                    Handle_server_set_get (context, from_client_set_get_cnt, all_synched, temp_degC);
                } break;
                // ----------------------------------------------------------------------------
                // PHASE II: all clients will poll here until ALL clients have done PHASE I
                //           Observe that these clients don't know of each other
                //           The clients poll here because we have found no way to make a
                //           notification type interface work for 3 clients using each server
                // ----------------------------------------------------------------------------
                case i_conn_0.poll_all_synched (void) -> all_clients_synched_t all_in_synch_return  : {
                    Handle_server_poll_all_clients_seen (
                            all_synched, from_client_poll_all_synched_cnt, all_in_synch_return);
                } break;
                case i_conn_1.poll_all_synched (void) -> all_clients_synched_t all_in_synch_return  : {
                    Handle_server_poll_all_clients_seen (
                            all_synched, from_client_poll_all_synched_cnt, all_in_synch_return);
                } break;
                case i_conn_2.poll_all_synched (void) -> all_clients_synched_t all_in_synch_return  : {
                    Handle_server_poll_all_clients_seen (
                            all_synched, from_client_poll_all_synched_cnt, all_in_synch_return);
                } break;
            }
        }
    } // Server_node_task

    int main (void) {
        #if ((NUM_ROWS==2) and (NUM_COLS==2))
            // TOP_24_HAS_1_TILE_004_T0_2N_2CN_ETC_MAYBE_CONVERGING
            /*
            Constraint check for tile[0]:
              Cores available:            8,   used:          3 .  OKAY
              Timers available:          10,   used:          3 .  OKAY
              Chanends available:        32,   used:          8 .  OKAY
              Memory available:       262144,   used:      45984 .  OKAY 45864
                (Stack: 15300, Code: 27888, Data: 2796)
            Constraints checks PASSED.
            Constraint check for tile[1]:
              Cores available:            8,   used:          1 .  OKAY
              Timers available:          10,   used:          1 .  OKAY
              Chanends available:        32,   used:          0 .  OKAY
              Memory available:       262144,   used:       1072 .  OKAY
                (Stack: 348, Code: 596, Data: 128)
            Constraints checks PASSED.
            */
            #if (TOP_WARNINGS==1) // From "makefile"
                #warning TOP_STR "Top_24"
            #endif

            conn_if_t conn [NUM_ROWS][NUM_PAIRS_PER_ROW][NUM_CONNS_PER_NODE]; // [ROW][PAIR][CONN] = [R][P][C]
                //         [    2   ][         1       ][        3         ]

            #define VER_NS_X [1][0][1] // NS      North-South top-bottom
            #define HOR_1_EW [1][0][2] // EW      EastWest belt side-to-side
            #define HOR_1_A  [1][0][0] // [A,B,C] Between horisontal columns
            #define VER_01_X [0][0][1] // [X,Y,Z] Between vertical rows
            #define HOR_0_EW [0][0][2] // EW      EastWest belt side-to-side
            #define HOR_0_A  [0][0][0] // [A,B,C] Between horisontal columns

            #if (TOP_24_AS_PARTLY_COMBINED==0)
                par {
                    Client_node_task (0,0, conn HOR_0_EW, conn VER_NS_X, conn HOR_0_A, outP4_leds);
                    Server_node_task (0,1, conn HOR_0_A,  conn VER_01_X, conn HOR_0_EW);
                    Client_node_task (1,0, conn HOR_1_EW, conn VER_01_X, conn HOR_1_A, null);
                    Server_node_task (1,1, conn HOR_1_A,  conn VER_NS_X, conn HOR_1_EW);
                }
            #elif (TOP_24_AS_PARTLY_COMBINED==1)

                #if (WARNINGS==1) // From "makefile"
                    #warning Does not work, see 13Mar2021 016
                #endif
                par {
                    on tile[0]: // (*)
                    par {
                        Client_node_task (0,0, conn HOR_0_EW, conn VER_NS_X, conn HOR_0_A, outP4_leds);
                        Server_node_task      (0,1, conn HOR_0_A,  conn VER_01_X, conn HOR_0_EW);
                    }
                    on tile[0]: // (*)
                    [[combine]]
                    par {
                        Client_node_task (1,0, conn HOR_1_EW, conn VER_01_X, conn HOR_1_A, null);
                        Server_node_task (1,1, conn HOR_1_A,  conn VER_NS_X, conn HOR_1_EW); // XMOS: LOCAL VALUES HERE ARE COMPILED WRONGLY! It's not printf!
                    }
                }
                // Avoids chanends 1+. MAYBE PASSED WITH CAVEATS etc. See versions 012 and 013

            #elif (TOP_24_AS_PARTLY_COMBINED==2)
                par {
                    on tile[0]: // (*)
                    par {
                        Client_node_task (0,0, conn HOR_0_EW, conn VER_NS_X, conn HOR_0_A, outP4_leds);
                        Server_node_task (1,1, conn HOR_1_A,  conn VER_NS_X, conn HOR_1_EW);
                    }
                    on tile[0]: // (*)
                    [[combine]]
                    par {
                        Client_node_task (1,0, conn HOR_1_EW, conn VER_01_X, conn HOR_1_A, null);
                        Server_node_task (0,1, conn HOR_0_A,  conn VER_01_X, conn HOR_0_EW);
                    }
                }
                // Avoids chanends 1+. MAYBE PASSED WITH CAVEATS etc. See versions 012 and 013
            #endif

        #elif ((NUM_ROWS==4) and (NUM_COLS==4))
            #if (TOP_WARNINGS==1) // From "makefile"
                #warning TOP_STR "Top_25"
            #endif
            /*
            Constraint check for tile[0]:
              Cores available:            8,   used:          3 .  OKAY
              Timers available:          10,   used:          3 .  OKAY
              Chanends available:        32,   used:          8 .  OKAY
              Memory available:       262144,   used:      55664 .  OKAY
                (Stack: 22140, Code: 30212, Data: 3312)
            Constraints checks PASSED.
            Constraint check for tile[1]:
              Cores available:            8,   used:          1 .  OKAY
              Timers available:          10,   used:          1 .  OKAY
              Chanends available:        32,   used:          0 .  OKAY
              Memory available:       262144,   used:       1072 .  OKAY
                (Stack: 348, Code: 596, Data: 128)
            Constraints checks PASSED.
             */
            conn_if_t conn [NUM_ROWS][NUM_PAIRS_PER_ROW][NUM_CONNS_PER_NODE]; // [ROW][PAIR][CONN] = [R][P][C]
                //         [    4   ][         2       ][        3         ]

            #define VER_NS_Y [3][1][1] // NS      North-South top-bottom
            #define VER_NS_X [3][0][1] // NS      North-South top-bottom

            #define HOR_3_EW [3][1][2] // EW      EastWest belt side-to-side
            #define HOR_3_C  [3][1][0] // [A,B,C] Between horisontal columns
            #define HOR_3_B  [3][0][2] // [A,B,C] Between horisontal columns
            #define HOR_3_A  [3][0][0] // [A,B,C] Between horisontal columns

            #define VER_23_Y [2][1][1] // [X,Y,Z] Between vertical rows
            #define VER_23_X [2][0][1] // [X,Y,Z] Between vertical rows
            #define HOR_2_EW [2][1][2] // EW      EastWest belt side-to-side
            #define HOR_2_C  [2][1][0] // [A,B,C] Between horisontal columns
            #define HOR_2_B  [2][0][2] // [A,B,C] Between horisontal columns
            #define HOR_2_A  [2][0][0] // [A,B,C] Between horisontal columns

            #define VER_12_Y [1][1][1] // [X,Y,Z] Between vertical rows
            #define VER_12_X [1][0][1] // [X,Y,Z] Between vertical rows
            #define HOR_1_EW [1][1][2] // EW      EastWest belt side-to-side
            #define HOR_1_C  [1][1][0] // [A,B,C] Between horisontal columns
            #define HOR_1_B  [1][0][2] // [A,B,C] Between horisontal columns
            #define HOR_1_A  [1][0][0] // [A,B,C] Between horisontal columns

            #define VER_01_Y [0][1][1] // [X,Y,Z] Between vertical rows
            #define VER_01_X [0][0][1] // [X,Y,Z] Between vertical rows
            #define HOR_0_EW [0][1][2] // EW      EastWest belt side-to-side
            #define HOR_0_C  [0][1][0] // [A,B,C] Between horisontal columns
            #define HOR_0_B  [0][0][2] // [A,B,C] Between horisontal columns
            #define HOR_0_A  [0][0][0] // [A,B,C] Between horisontal columns

            #define DO_COMBINED 1

            #if (DO_COMBINED == 1)
                par {
                    on tile[0]:
                    par {
                        Client_node_task (0,0, conn HOR_0_EW, conn VER_NS_X, conn HOR_0_A, outP4_leds);
                        Server_node_task (0,1, conn HOR_0_A,  conn VER_01_X, conn HOR_0_B);
                    }
                    on tile[0]:
                    [[combine]]
                    par {
                        Client_node_task (0,2, conn HOR_0_B,  conn VER_NS_Y, conn HOR_0_C, null);
                        Server_node_task (0,3, conn HOR_0_C,  conn VER_01_Y, conn HOR_0_EW);

                        Client_node_task (1,0, conn HOR_1_EW, conn VER_01_X, conn HOR_1_A, null);
                        Server_node_task (1,1, conn HOR_1_A,  conn VER_12_X, conn HOR_1_B);
                        Client_node_task (1,2, conn HOR_1_B,  conn VER_01_Y, conn HOR_1_C, null);
                        Server_node_task (1,3, conn HOR_1_C,  conn VER_12_Y, conn HOR_1_EW);

                        Client_node_task (2,0, conn HOR_2_EW, conn VER_12_X, conn HOR_2_A, null);
                        Server_node_task (2,1, conn HOR_2_A,  conn VER_23_X, conn HOR_2_B);
                        Client_node_task (2,2, conn HOR_2_B,  conn VER_12_Y, conn HOR_2_C, null);
                        Server_node_task (2,3, conn HOR_2_C,  conn VER_23_Y, conn HOR_2_EW);

                        Client_node_task (3,0, conn HOR_3_EW, conn VER_23_X, conn HOR_3_A, null);
                        Server_node_task (3,1, conn HOR_3_A,  conn VER_NS_X, conn HOR_3_B);
                        Client_node_task (3,2, conn HOR_3_B,  conn VER_23_Y, conn HOR_3_C, null);
                        Server_node_task (3,3, conn HOR_3_C,  conn VER_NS_Y, conn HOR_3_EW);
                    }
                }
            #elif (DO_COMBINED == 0)
                par {
                    on tile[0]:
                    par {
                        Client_node_task (0,0, conn HOR_0_EW, conn VER_NS_X, conn HOR_0_A, outP4_leds);
                        Server_node_task (0,1, conn HOR_0_A,  conn VER_01_X, conn HOR_0_B);
                        Client_node_task (0,2, conn HOR_0_B,  conn VER_NS_Y, conn HOR_0_C, null);
                        Server_node_task (0,3, conn HOR_0_C,  conn VER_01_Y, conn HOR_0_EW);

                        Client_node_task (1,0, conn HOR_1_EW, conn VER_01_X, conn HOR_1_A, null);
                        Server_node_task (1,1, conn HOR_1_A,  conn VER_12_X, conn HOR_1_B);
                        Client_node_task (1,2, conn HOR_1_B,  conn VER_01_Y, conn HOR_1_C, null);
                        Server_node_task (1,3, conn HOR_1_C,  conn VER_12_Y, conn HOR_1_EW);
                    }
                    on tile[1]:
                    par {
                        Client_node_task (2,0, conn HOR_2_EW, conn VER_12_X, conn HOR_2_A, null);
                        Server_node_task (2,1, conn HOR_2_A,  conn VER_23_X, conn HOR_2_B);
                        Client_node_task (2,2, conn HOR_2_B,  conn VER_12_Y, conn HOR_2_C, null);
                        Server_node_task (2,3, conn HOR_2_C,  conn VER_23_Y, conn HOR_2_EW);

                        Client_node_task (3,0, conn HOR_3_EW, conn VER_23_X, conn HOR_3_A, null);
                        Server_node_task (3,1, conn HOR_3_A,  conn VER_NS_X, conn HOR_3_B);
                        Client_node_task (3,2, conn HOR_3_B,  conn VER_23_Y, conn HOR_3_C, null);
                        Server_node_task (3,3, conn HOR_3_C,  conn VER_NS_Y, conn HOR_3_EW);
                    }
                }
            #endif

        #endif

        return 0;

    } // main

#endif /* TOP_24_XC_H_ */
